import logging
import requests
import json
from operator import itemgetter
import urllib.parse

logger = logging.getLogger(__name__)

def add_version_as_component(self, main_project_release, sub_project_release):
    headers = self.get_headers()
    main_data = main_project_release['_meta']['href'].split('/')
    sub_data = sub_project_release['_meta']['href'].split('/')
    main_project_release_links = main_project_release['_meta']['links']
    main_project_release_component_links = [x for x in main_project_release_links if x['rel'] == 'components']
    main_project_release_component_link = main_project_release_component_links[0]['href']
    logger.debug(main_project_release_component_link)
    sub_project_release_as_custom_component_url = self.get_apibase() + "/components/" + sub_data[5] + "/versions/" + sub_data[7]
    logger.debug(sub_project_release_as_custom_component_url)
    payload = {}
    payload['component'] = sub_project_release_as_custom_component_url
    logger.debug(json.dumps(payload))
    response = requests.post(main_project_release_component_link, headers=headers, verify = not self.config['insecure'], json=payload)
    logger.debug(response)
    return response



def remove_version_as_component(self, main_project_release, sub_project_release):
    headers = self.get_headers()
    main_data = main_project_release['_meta']['href'].split('/')
    sub_data = sub_project_release['_meta']['href'].split('/')
    main_project_release_links = main_project_release['_meta']['links']
    main_project_release_component_links = [x for x in main_project_release_links if x['rel'] == 'components']
    main_project_release_component_link = main_project_release_component_links[0]['href']
    logger.debug(main_project_release_component_link)
    subcomponent_url = main_project_release_component_link + "/" + sub_data[5] + "/versions/" + sub_data[7]
    logger.debug(subcomponent_url)
    response = requests.delete(subcomponent_url, headers=headers, verify = not self.config['insecure'])
    return response



import logging
import requests
import json
from operator import itemgetter
import urllib.parse
from .Exceptions import UnsupportedBDVersion

logger = logging.getLogger(__name__)

def get_file_matches_for_bom_component(self, bom_component, limit=1000):
    self._check_version_compatibility()
    url = self.get_link(bom_component, "matched-files")
    paramstring = self.get_limit_paramstring(limit)
    logger.debug("GET {}".format(url))
    response = self.execute_get(url)
    jsondata = response.json()
    return jsondata




import logging
import requests
import json
from operator import itemgetter
import urllib.parse

from .Exceptions import UnknownVersion, CreateFailedAlreadyExists, CreateFailedUnknown

logger = logging.getLogger(__name__)

def read_config(self):
    try:
        with open('.restconfig.json','r') as f:
            self.config = json.load(f)
    except:
        logging.error(f"Unable to load configuration from '.restconfig.json'. Make sure you create one with proper connection and authentication values for your Black Duck server")
        raise
        
def write_config(self):
    with open(self.configfile,'w') as f:
        json.dump(self.config, f, indent=3)
        
def get_auth_token(self):
    api_token = self.config.get('api_token', False)
    if api_token:
        authendpoint = "/api/tokens/authenticate"
        url = self.config['baseurl'] + authendpoint
        session = requests.session()
        response = session.post(
            url, 
            data={}, 
            headers={'Authorization': 'token {}'.format(api_token)}, 
            verify=not self.config['insecure']
        )
        csrf_token = response.headers['X-CSRF-TOKEN']
        try:
            bearer_token = json.loads(response.content.decode('utf-8'))['bearerToken']
        except json.decoder.JSONDecodeError as e:
            logger.exception("Authentication failure, could not obtain bearer token")
            raise Exception("Failed to obtain bearer token, check for valid authentication token")
        return (bearer_token, csrf_token, None)
    else:
        authendpoint="/j_spring_security_check"
        url = self.config['baseurl'] + authendpoint
        session=requests.session()
        credentials = dict()
        credentials['j_username'] = self.config['username']
        credentials['j_password'] = self.config['password']
        response = session.post(url, credentials, verify= not self.config['insecure'])
        cookie = response.headers['Set-Cookie']
        token = cookie[cookie.index('=')+1:cookie.index(';')]
    return (token, None, cookie)

def _get_hub_rest_api_version_info(self):
    '''Get the version info from the server, if available
    '''
    session = requests.session()
    url = self.config['baseurl'] + "/api/current-version"
    response = session.get(url, verify = not self.config['insecure'])

    if response.status_code == 200:
        version_info = response.json()
        if 'version' in version_info:
            return version_info
        else:
            raise UnknownVersion("Did not find the 'version' key in the response to a successful GET on /api/current-version")
    else:
        raise UnknownVersion("Failed to retrieve the version info from {}, status code {}".format(url, response.status_code))

def _get_major_version(self):
    return self.version_info['version'].split(".")[0]

def get_urlbase(self):
    return self.config['baseurl']

def get_headers(self):
    if self.config.get('api_token', False):
        return {
            'X-CSRF-TOKEN': self.csrf_token, 
            'Authorization': 'Bearer {}'.format(self.token), 
            'Accept': 'application/json',
            'Content-Type': 'application/json'}
    else:
        if self.bd_major_version == "3":
            return {"Cookie": self.cookie}
        else:
            return {"Authorization":"Bearer " + self.token}

def get_api_version(self):
    url = self.get_urlbase() + '/api/current-version'
    response = self.execute_get(url)
    version = response.json().get('version', 'unknown')
    return version

def _get_parameter_string(self, parameters={}):
    parameter_string = "&".join(["{}={}".format(k,urllib.parse.quote(str(v))) for k,v in sorted(parameters.items(), key=itemgetter(0))])
    return "?" + parameter_string

def get_tags_url(self, component_or_project):
    # Utility method to return the tags URL from either a component or project object
    url = None
    for link_d in component_or_project['_meta']['links']:
        if link_d['rel'] == 'tags':
            return link_d['href']
    return url

def get_link(self, bd_rest_obj, link_name):
    # returns the URL for the link_name OR None
    if bd_rest_obj and '_meta' in bd_rest_obj and 'links' in bd_rest_obj['_meta']:
        for link_obj in bd_rest_obj['_meta']['links']:
            if 'rel' in link_obj and link_obj['rel'] == link_name:
                return link_obj.get('href', None)
    else:
        logger.warning("This does not appear to be a BD REST object. It should have ['_meta']['links']")

def get_limit_paramstring(self, limit):
    return "?limit={}".format(limit)

def get_apibase(self):
    return self.config['baseurl'] + "/api"

def execute_delete(self, url):
    headers = self.get_headers()
    response = requests.delete(url, headers=headers, verify = not self.config['insecure'])
    return response

def _validated_json_data(self, data_to_validate):
    if isinstance(data_to_validate, dict) or isinstance(data_to_validate, list):
        json_data = json.dumps(data_to_validate)
    else:
        json_data = data_to_validate
    json.loads(json_data) # will fail with JSONDecodeError if invalid
    return json_data

def execute_get(self, url, custom_headers={}):
    headers = self.get_headers()
    headers.update(custom_headers)
    response = requests.get(url, headers=headers, verify = not self.config['insecure'])
    return response
    
def execute_put(self, url, data, custom_headers={}):
    json_data = self._validated_json_data(data)
    headers = self.get_headers()
    headers["Content-Type"] = "application/json"
    headers.update(custom_headers)
    response = requests.put(url, headers=headers, data=json_data, verify = not self.config['insecure'])
    return response

def _create(self, url, json_body):
    response = self.execute_post(url, json_body)
    # v4+ returns the newly created location in the response headers
    # and there is nothing in the response json
    # whereas v3 returns the newly created object in the response json
    if response.status_code == 201:
        if "location" in response.headers:
            return response.headers["location"]
        else:
            try:
                response_json = response.json()
            except json.decoder.JSONDecodeError:
                logger.warning('did not receive any json data back')
            else:
                if '_meta' in response_json and 'href' in response_json['_meta']:
                    return response_json['_meta']['href']
                else:
                    return response_json
    elif response.status_code == 412:
        raise CreateFailedAlreadyExists("Failed to create the object because it already exists - url {}, body {}, response {}".format(url, json_body, response))
    else:
        raise CreateFailedUnknown("Failed to create the object for an unknown reason - url {}, body {}, response {}".format(url, json_body, response))

def execute_post(self, url, data, custom_headers={}):
    json_data = self._validated_json_data(data)
    headers = self.get_headers()
    headers["Content-Type"] = "application/json"
    headers.update(custom_headers)
    response = requests.post(url, headers=headers, data=json_data, verify = not self.config['insecure'])
    return response

def get_matched_components(self, version_obj, limit=9999):
    url = "{}/matched-files".format(version_obj['_meta']['href'])
    param_string = self._get_parameter_string({'limit': limit})
    url = "{}{}".format(url, param_string)
    response = self.execute_get(url)
    return response.json()

def _check_version_compatibility(self):
    if int(self.bd_major_version) < 2018:
        raise UnsupportedBDVersion("The BD major version {} is less than the minimum required major version {}".format(self.bd_major_version, 2018))        

    



import logging
import requests
import json
from operator import itemgetter
import urllib.parse

logger = logging.getLogger(__name__)

def get_jobs(self, parameters={}):
    url = self.get_apibase() + "/jobs"
    url = url + self._get_parameter_string(parameters)
    custom_headers = {'Accept': 'application/vnd.blackducksoftware.status-4+json'}
    response = self.execute_get(url, custom_headers=custom_headers)
    return response.json()

def get_job_statistics(self):
    url = self.get_urlbase() + "/api/job-statistics"
    response = self.execute_get(url)
    return response.json()
    


import logging
import requests
import json
from operator import itemgetter
import urllib.parse

logger = logging.getLogger(__name__)

def _get_user_group_url(self):
    return self.config['baseurl'] + "/api/usergroups"

def get_user_groups(self, parameters={}):
    url = self._get_user_group_url() + self._get_parameter_string(parameters)
    headers = {'Accept': 'application/vnd.blackducksoftware.user-4+json'}
    response = self.execute_get(url, custom_headers=headers)
    return response.json()

def get_user_group_by_name(self, group_name):
    group_list = self.get_user_groups({"q": f"name:{group_name}"})
    for group in group_list['items']:
        if group['name'] == group_name:
            return group

def create_user_group(self, user_group_json):
    if self.bd_major_version == "3":
        url = self.config['baseurl'] + '/api/v1/usergroups'
    else:
        url = self._get_user_group_url()
    location = self._create(url, user_group_json)
    return location

def create_user_group_by_name(self, group_name, active=True):
    user_group_info = {
        'name': group_name,
        'createdFrom': 'INTERNAL',
        'active': active
    }
    return self.create_user_group(user_group_info)

# def get_user_group_by_id(self, user_group_id):
#     url = self._get_user_group_url() + "/{}".format(user_group_id)
#     return self.get_user_group_by_url(url)

# def get_user_group_by_url(self, user_group_url):
#     response = self.execute_get(user_group_url)
#     jsondata = response.json()
#     return jsondata

# def get_user_group_by_name(self, user_group_name):
#     url = self._get_user_group_url() + "?q={}".format(user_group_name)
#     response = self.execute_get(url)
#     user_group_obj = response.json()
#     if user_group_obj['totalCount'] > 0:
#         return user_group_obj['items'][0]

def update_user_group_by_id(self, user_group_id, update_json):
    url = self._get_user_group_url() + "/{}".format(user_group_id)
    return self.update_user_group_by_url(url, update_json)

def update_user_group_by_url(self, user_group_url, update_json):
    return self.execute_put(user_group_url, update_json)

def delete_user_group_by_id(self, user_group_id):
    url = self._get_user_group_url() + "/{}".format(user_group_id)
    return self.delete_user_group_by_url(url)

def delete_user_group_by_url(self, user_group_url):
    return self.execute_delete(user_group_url)



import logging
import requests
import json
from operator import itemgetter
import urllib.parse

logger = logging.getLogger(__name__)

def get_licenses(self, parameters={}):
    url = self.get_urlbase() + "/api/licenses" + self._get_parameter_string(parameters)
    response = self.execute_get(url, custom_headers={'Accept':'application/json'})
    json_data = response.json()
    return json_data

def _get_license_info(self, license_obj):
    if 'license' in license_obj:
        license_info = {}
        text_json = {}
        logger.debug("license: {}".format(license_obj))
        response = self.execute_get(license_obj['license'])
        if response.status_code == 200:
            license_info = response.json()
            text_url = self.get_link(license_info, 'text')
            response = self.execute_get(text_url)
            if response.status_code == 200:
                text_json = response.text
        yield {"license_info": license_info,
                "license_text_info": text_json}
    elif 'licenses' in license_obj and isinstance(license_obj['licenses'], list):
        for license in license_obj['licenses']:
            self._get_license_info(license)

def get_license_info_for_bom_component(self, bom_component, limit=1000):
    self._check_version_compatibility()
    all_licenses = {}
    logger.debug("gathering license info for bom component {}, version {}".format(
        bom_component['componentName'], bom_component['componentVersionName']))
    for license in bom_component.get('licenses', []):
        for license_info_obj in self._get_license_info(license):
            all_licenses.update({
                    license['licenseDisplay']: license_info_obj
                })
    return all_licenses





'''
Created on Jul 6, 2018

@author: kumykov
@contributors: gsnyder2007, AR-Calder

Wrapper for common HUB API queries. 
Upon initialization Bearer tocken is obtained and used for all subsequent calls

Usage: 

credentials and hub URL could be placed in the .restconfig.json file
    
    {
      "baseurl": "https://hub-hostname",
      "username": "<username goes here>",
      "password": "<password goes here>",
      "insecure": true,
      "debug": false
    }

    OR, using API Token

    {
      "baseurl": "https://hub-hostname",
      "api_token": "<API token goes here>",
      "insecure": true,
      "debug": false
    }

    .restconfig.json should be present in the current directory.
    
    from blackduck.HubRestApi import HubInstance
    
    hub = HubInstance()
    projects = hub.get_projects()

It is possible to generate generate_config file by initalizing API as following:
   
    from blackduck.HubRestApi import HubInstance
    
    username="<username goes here>"
    password="<password goes here>"
    urlbase="https://hub-hostname"
    
    hub = HubInstance(urlbase, username, password, insecure=True)
    
    
'''
import logging
import requests

logger = logging.getLogger(__name__)

from .Utils import object_id

class HubInstance(object):
    '''
    classdocs
    '''

    from .constants import VERSION_DISTRIBUTION, VERSION_PHASES, PROJECT_VERSION_SETTINGS

    # TODO: What to do about the config file for thread-safety, concurrency
    configfile = ".restconfig.json"
      
    from .Core import (
        _create,_get_hub_rest_api_version_info,_get_major_version,_get_parameter_string,_validated_json_data,
        execute_delete,execute_get,execute_post,execute_put,get_api_version,get_apibase,get_auth_token,get_headers,
        get_limit_paramstring,get_link,get_matched_components,get_tags_url,get_urlbase,read_config,write_config,
        _check_version_compatibility
    )
    from .Roles import (
        _get_role_url, assign_role_given_role_url, assign_role_to_user_or_group, 
        delete_role_from_user_or_group, get_role_url_by_name, get_roles, get_roles_for_user_or_group,
        get_roles_url_from_user_or_group, user_has_role
    )
    from .Users import (
        _get_user_url, create_user, delete_user_by_id, delete_user_by_url, get_current_user,
        get_last_login, get_user_by_id, get_user_by_url, get_users, reset_user_password, update_user_by_id, 
        update_user_by_url
    )
    from .UserGroup import (
        _get_user_group_url, create_user_group, create_user_group_by_name, 
        delete_user_group_by_id, delete_user_group_by_url, get_user_group_by_name, get_user_groups, 
        update_user_group_by_id, update_user_group_by_url
    )
    from .Policy import (
        _get_policy_url, create_policy, delete_policy_by_id, delete_policy_by_url, 
        get_policies, get_policy_by_id, get_policy_by_url, update_policy_by_id, update_policy_by_url
    )
    from .Vulnerabilities import (
        _get_vulnerabilities_url, get_component_remediation, get_vulnerabilities, 
        get_vulnerability_affected_projects, get_vulnerable_bom_components
    )
    from .Reporting import (
        create_version_notices_report, create_version_reports, create_vuln_status_report, 
        download_notification_report, download_report
    )
    from .Projects import (
        _find_user_group_url, _find_user_url, _get_projects_url, _project_role_urls, 
        assign_project_application_id, assign_user_group_to_project, assign_user_to_project, 
        compare_project_versions, create_project, create_project_version, delete_all_empty_versions, 
        delete_application_id, delete_empty_projects, delete_empty_versions, delete_project_by_name, 
        delete_project_version_by_name, delete_project_version_codelocations, delete_user_group_from_project, 
        get_or_create_project_version, get_project_application_id, get_project_by_id, get_project_by_name, 
        get_project_info, get_project_roles, get_project_version_by_name, get_project_versions, get_projects, 
        get_projects_by_version_name, get_version_by_id, get_version_by_name, get_version_codelocations, 
        get_version_components, get_version_scan_info, update_project_application_id, update_project_settings, 
        update_project_version_settings
    ) # TODO Transfer relevant versions related functions to .Versions
    from .Versions import ( add_version_as_component, remove_version_as_component )
    from .Scans import (
        delete_codelocation, delete_unmapped_codelocations, download_project_scans, 
        get_codelocation_scan_summaries, get_codelocations, get_codelocations_internal, get_scan_locations, upload_scan
    )
    from .Components import (
        _get_components_url, find_component_info_for_protex_component, get_component_by_id, 
        get_component_by_url, get_components, search_components, update_component_by_id, update_component_by_url
    )
    from .CustomFields import (
        _get_cf_obj_rel_path, _get_cf_object_url, _get_cf_url, create_cf, delete_cf, 
        get_cf_object, get_cf_objects, get_cf_value, get_cf_values, get_custom_fields, put_cf_value,
        supported_cf_object_types
    )
    from .Licences import ( _get_license_info, get_license_info_for_bom_component, get_licenses )
    from .Snippet import ( get_file_matches_for_bom_component )
    from .System import ( get_health_checks, get_notifications )
    from .Ldap import ( disable_ldap, enable_ldap, get_ldap_configs, get_ldap_state )

    def __init__(self, *args, **kwargs):
        # Config needs to be an instance variable for thread-safety, concurrent use of HubInstance()
        self.config = {}

        try:
            self.config['baseurl'] = args[0]
            api_token = kwargs.get('api_token', False)
            if api_token:
                self.config['api_token'] = api_token
            else:
                self.config['username'] = args[1]
                self.config['password'] = args[2]
            self.config['insecure'] = kwargs.get('insecure', False)
            self.config['debug'] = kwargs.get('debug', False)

            if kwargs.get('write_config_flag', True):
                self.write_config()
        except Exception:
            self.read_config()
            
        if self.config['insecure']:
            requests.packages.urllib3.disable_warnings()
        
        if self.config['debug']:
            logger.debug(f"Reading connection and authentication info from {self.configfile}")
        
        self.token, self.csrf_token, self.cookie = self.get_auth_token()
        try:
            self.version_info = self._get_hub_rest_api_version_info()
        except UnknownVersion:
            self.version_info = {'version': '3'} # assume it's v3 since all versions after 3 supported version info

        self.bd_major_version = self._get_major_version()

    def print_methods(self):
        import inspect
        for fn in inspect.getmembers(self, predicate=inspect.isfunction):
            print(fn[0])



import logging
import requests
import json
from operator import itemgetter
import urllib.parse

logger = logging.getLogger(__name__)

def _get_role_url(self):
    return self.config['baseurl'] + "/api/roles"

def get_roles(self, parameters={}):
    url = self._get_role_url() + self._get_parameter_string(parameters)
    response = self.execute_get(url)
    return response.json()  

def get_roles_url_from_user_or_group(self, user_or_group):
    # Given a user or user group object, return the 'roles' url
    roles_url = None
    for endpoint in user_or_group['_meta']['links']:
        if endpoint['rel'] == "roles":
            roles_url = endpoint['href']
    return roles_url

def get_roles_for_user_or_group(self, user_or_group):
    roles_url = self.get_roles_url_from_user_or_group(user_or_group)
    if roles_url:
        response = self.execute_get(roles_url)
        return response.json()
    else:
        return []

def get_role_url_by_name(self, role_name):
    # Return the global (as opposed to project-specific) role URL for this server corresponding to the role name
    all_roles = self.get_roles()
    for role in all_roles['items']:
        if role['name'] == role_name:
            return role['_meta']['href']

def assign_role_to_user_or_group(self, role_name, user_or_group):
    user_or_group_roles_url = self.get_roles_url_from_user_or_group(user_or_group)
    return self.assign_role_given_role_url(role_name, user_or_group_roles_url)

def assign_role_given_role_url(self, role_name, user_or_group_role_assignment_url):
    role_url = self.get_role_url_by_name(role_name)
    if self.bd_major_version == "3":
        # A hack to get the assignment to work on v3
        role_url = role_url.replace("api", "api/internal")
    data = {"name": role_name, "role": role_url}
    logger.debug("executing POST to {} with {}".format(
        user_or_group_role_assignment_url, data))
    return self.execute_post(user_or_group_role_assignment_url, data = data)

def delete_role_from_user_or_group(self, role_name, user_or_group):
    roles = self.get_roles_for_user_or_group(user_or_group)
    for role in roles['items']:
        if role['name'] == role_name:
            self.execute_delete(role['_meta']['href'])


def user_has_role(self, user_or_group, role_name):
    user_roles_obj = self.get_roles_for_user_or_group(user_or_group)
    return role_name in [r['name'] for r in user_roles_obj['items']]



"""
Created on Dec 23, 2020
@author: ar-calder
"""
 
import requests
from requests.auth import AuthBase
import logging
import json
from datetime import datetime, timedelta

logger = logging.getLogger(__name__)


class NoAuth(AuthBase):
    def __call__(self, r):
        return r


class BearerAuth(AuthBase):
    """Authenticate with Blackduck hub using access token"""

    def __init__(self, session, token):
        """
        Args:
            session (requests.session): requests session to authenticate
            token (string): of Blackduck user from UI: System -> My Access Tokens
        """
        if any(arg is False for arg in (session, token)):
            raise ValueError(
                'session & token are required'
            )

        self.session = session
        self.access_token = token
        self.bearer_token = None
        self.csrf_token = None
        self.valid_until = datetime.now()

    def __call__(self, request):
        if not self.bearer_token or datetime.now() > self.valid_until - timedelta(minutes=5):
            # If bearer token not set or nearing expiry
            self.authenticate()

        request.headers.update({ 
            "authorization": f"bearer {self.bearer_token}",
            "X-CSRF-TOKEN": self.csrf_token
        })

        return request

    def authenticate(self):
        if not self.session.verify:
            requests.packages.urllib3.disable_warnings()
            # Announce this on every auth attempt, as a little incentive to properly configure certs
            logger.warning("ssl verification disabled, connection insecure. do NOT use verify=False in production!")

        response = self.session.post(
            url="/api/tokens/authenticate",
            auth=NoAuth(),  # temporarily strip authentication to avoid infinite recursion
            headers={"Authorization": f"token {self.access_token}"}
        )

        if response.status_code == 200:
            try:
                content = response.json()
                self.bearer_token = content['bearerToken']
                self.csrf_token = response.headers['X-CSRF-TOKEN']
                self.valid_until = datetime.now() + timedelta(milliseconds=int(content['expiresInMilliseconds']))
                logger.info(f"success: auth granted until {self.valid_until.astimezone()}")
                return
            except (json.JSONDecodeError, KeyError):
                logger.exception("HTTP response status code 200 but unable to obtain bearer token")
                # fall through

        if response.status_code == 401:
            logger.error("HTTP response status code = 401 (Unauthorized)")
            try:
                logger.error(response.json()['errorMessage'])
            except (json.JSONDecodeError, KeyError):
                logger.exception("unable to extract error message")
                logger.error("HTTP response headers: %s", response.headers)
                logger.error("HTTP response text: %s", response.text)
            raise RuntimeError("Unauthorized access token", response)

        # all unhandled responses fall through to here
        logger.error("Unhandled HTTP response")
        logger.error("HTTP response status code %i", response.status_code)
        logger.error("HTTP response headers: %s", response.headers)
        logger.error("HTTP response text: %s", response.text)
        raise RuntimeError("Unhandled HTTP response", response)


class CookieAuth(AuthBase):
    """Authenticate with Blackduck hub using username/password

       Note: username/password is not recommended and Client users are encouraged
             to use the more secure token authentication instead.
    """

    def __init__(self, session, username, password):
        """
        Args:
            session (requests.session): requests session to authenticate
            username (string): of Blackduck hub user
            password (string): of Blackduck hub user
        """
        if any(arg is False for arg in (session, username, password)):
            raise ValueError(
                'session, username and password are required'
            )

        self.session = session
        self.username = username
        self.password = password
        self.bearer_token = None
        self.csrf_token = None
        self.valid_until = datetime.now()

    def __call__(self, request):
        if not self.bearer_token or datetime.now() > self.valid_until - timedelta(minutes=5):
            # If bearer token not set or nearing expiry
            self.authenticate()

        request.headers.update({ 
            "authorization": f"bearer {self.bearer_token}",
            "X-CSRF-TOKEN": self.csrf_token
        })

        return request

    def authenticate(self):
        logger.warning("Authenticating with username/password is not recommended. Consider using the more secure "
                       "token based authentication instead.")
        if not self.session.verify:
            requests.packages.urllib3.disable_warnings()
            # Announce this on every auth attempt, as a little incentive to properly configure certs
            logger.warning("ssl verification disabled, connection insecure. do NOT use verify=False in production!")

        credentials = {'j_username': self.username, 'j_password': self.password}
        response = self.session.post(
            url="/j_spring_security_check",
            data=credentials,
            auth=NoAuth()  # temporarily strip authentication to avoid infinite recursion
        )

        if response.status_code == 204:  # No Content
            try:
                cookie = response.headers['Set-Cookie']
                self.bearer_token = cookie[cookie.index('=') + 1:cookie.index(';')]
                self.csrf_token = response.headers['X-CSRF-TOKEN']
                # As of 2021.2 the bearer token is good for 2 hours but there
                # is no explicit reference to expiry time in the response.
                #
                # There is internal talk (Feb. 2021) of revamping authentication and the
                # validity time will likely be reduced or the /j_spring_security_check
                # may be deprecated.
                #
                # HUB-25720: It is not possible to extend the validity time
                # of the bearer token obtained via /j_spring_security_check.
                self.valid_until = datetime.now() + timedelta(minutes=120)  # token is good for 2 hours
                logger.info(f"success: auth granted until {self.valid_until.astimezone()}")
                return
            except (KeyError, ValueError):
                logger.exception("HTTP response status code 204 but unable to obtain bearer token")
                # fall through

        if response.status_code == 401:
            logger.error("HTTP response status code = 401 (Unauthorized)")
            try:
                logger.error(response.json()['errorMessage'])
            except (json.JSONDecodeError, KeyError):
                logger.exception("unable to extract error message")
                logger.error("HTTP response headers: %s", response.headers)
                logger.error("HTTP response text: %s", response.text)
            raise RuntimeError("Unauthorized username/password", response)

        # all unhandled responses fall through to here
        logger.error("Unhandled HTTP response")
        logger.error("HTTP response status code %i", response.status_code)
        logger.error("HTTP response headers: %s", response.headers)
        logger.error("HTTP response text: %s", response.text)
        raise RuntimeError("Unhandled HTTP response", response)



import logging
import requests
import json
from operator import itemgetter
import urllib.parse

logger = logging.getLogger(__name__)

def get_health_checks(self):
    url = self.get_urlbase() + "/api/health-checks/liveness"
    return self.execute_get(url)

def get_notifications(self, parameters={}):
    url = self.get_urlbase() + "/api/notifications" + self._get_parameter_string(parameters)
    custom_headers = {'Accept': 'application/vnd.blackducksoftware.notification-4+json'}
    response = self.execute_get(url, custom_headers=custom_headers)
    json_data = response.json()
    return json_data



#
# Black Duck Product Constants
#
VERSION_DISTRIBUTION=["EXTERNAL", "SAAS", "INTERNAL", "OPENSOURCE"]
VERSION_PHASES = ["PLANNING", "DEVELOPMENT", "PRERELEASE", "RELEASED", "DEPRECATED", "ARCHIVED"]
PROJECT_VERSION_SETTINGS = ['nickname', 'releaseComments', 'versionName', 'phase', 'distribution', 'releasedOn']



import logging
import requests
import json
from operator import itemgetter
import urllib.parse

logger = logging.getLogger(__name__)

def get_ldap_state(self):
    url = self.config['baseurl'] + "/api/v1/ldap/state"
    headers = self.get_headers()
    response = requests.get(url, headers=headers, verify = not self.config['insecure'])
    jsondata = response.json()
    return jsondata

def enable_ldap(self):
    url = self.config['baseurl'] + "/api/v1/ldap/state"
    headers = self.get_headers()
    payload = {}
    payload['ldapEnabled'] = True
    response = requests.post(url, headers=headers, verify = not self.config['insecure'], json=payload)
    jsondata = response.json()
    return jsondata
    
def disable_ldap(self):
    url = self.config['baseurl'] + "/api/v1/ldap/state"
    headers = self.get_headers()
    payload = {}
    payload['ldapEnabled'] = False
    response = requests.post(url, headers=headers, verify = not self.config['insecure'], json=payload)
    jsondata = response.json()
    return jsondata
    
def get_ldap_configs(self):
    url = self.config['baseurl'] + "/api/v1/ldap/configs"
    headers = self.get_headers()
    headers['Content-Type']  = "application/json"
    response = requests.get(url, headers=headers, verify = not self.config['insecure'])
    jsondata = response.json()
    return jsondata


import logging
import requests
import json
from operator import itemgetter
import urllib.parse

logger = logging.getLogger(__name__)

def _get_user_url(self):
    return self.config['baseurl'] + "/api/users"

def get_users(self, parameters={}):
    url = self._get_user_url() + self._get_parameter_string(parameters)
    headers = {'Accept': 'application/vnd.blackducksoftware.user-4+json'}
    response = self.execute_get(url, custom_headers=headers)
    return response.json()

def get_current_user(self):
    url = self.config['baseurl'] + "/api/current-user"
    headers = {'Accept': 'application/vnd.blackducksoftware.user-4+json'}
    response = self.execute_get(url, custom_headers=headers)
    return response.json()

def create_user(self, user_json):
    url = self._get_user_url()
    location = self._create(url, user_json)
    return location

def get_user_by_id(self, user_id):
    url = self._get_user_url() + "/{}".format(user_id)
    headers = {'Accept': 'application/vnd.blackducksoftware.user-4+json'}
    return self.get_user_by_url(url, custom_headers=headers)

def get_user_by_url(self, user_url):
    headers = {'Accept': 'application/vnd.blackducksoftware.user-4+json'}
    response = self.execute_get(user_url, custom_headers=headers)
    jsondata = response.json()
    return jsondata

def update_user_by_id(self, user_id, update_json):
    url = self._get_user_url() + "/{}".format(user_id)
    return self.update_user_by_url(url, update_json)

def update_user_by_url(self, user_url, update_json):
    return self.execute_put(user_url, update_json)

def delete_user_by_id(self, user_id):
    url = self._get_user_url() + "/{}".format(user_id)
    return self.delete_user_by_url(url)

def delete_user_by_url(self, user_url):
    return self.execute_delete(user_url)
    
def reset_user_password(self, user_id, new_password):
    url = self.config['baseurl'] + "/api/users/" + user_id + "/resetpassword"
    headers = {'Content-Type':'application/vnd.blackducksoftware.user-1+json', 'Accept': 'application/json'}
    data = {'password': new_password}
    return self.execute_put(url, data, headers)

def get_last_login(self,sinceDays=60):
    url = self.config['baseurl'] + "/api/dormant-users"
    param_string = self._get_parameter_string({'sinceDays': sinceDays})
    url = "{}{}".format(url, param_string)
    headers = {'Accept': 'application/vnd.blackducksoftware.user-4+json'}
    response = self.execute_get(url, custom_headers=headers)
    return response.json()
